@@ -226,6 +226,26 @@ private:
     memset(&fsid, 0, sizeof(fsid));
   }
 
+  // no copying
+  /* oh, how i long for c++11...
+private:
+  OSDMap(const OSDMap& other) = default;
+  const OSDMap& operator=(const OSDMap& other) = default;
+public:
+  */
+
+  void deepish_copy_from(const OSDMap& o) {
+    *this = o;
+    pg_temp.reset(new map<pg_t,vector<int> >(*o.pg_temp));
+    osd_uuid.reset(new vector<uuid_d>(*o.osd_uuid));
+
+    // NOTE: this still references shared entity_addr_t's.
+    osd_addrs.reset(new addrs_s(*o.osd_addrs));
+
+    // NOTE: we do not copy crush.  note that apply_incremental will
+    // allocate a new CrushWrapper, though.
+  }
+
   // map info
   const uuid_d& get_fsid() const { return fsid; }
   void set_fsid(uuid_d& f) { fsid = f; }
@@ -466,12 +486,13 @@ public:
     return object_locator_t(layout.fl_pg_pool);
   }
 
+  // XXX: not used, mentioned in psim.cc comment
   // oid -> pg
-  ceph_object_layout file_to_object_layout(object_t oid, ceph_file_layout& layout) const {
-    return make_object_layout(oid, layout.fl_pg_pool);
+  ceph_object_layout file_to_object_layout(object_t oid, ceph_file_layout& layout, string nspace) const {
+    return make_object_layout(oid, layout.fl_pg_pool, nspace);
   }
 
-  ceph_object_layout make_object_layout(object_t oid, int pg_pool) const;
+  ceph_object_layout make_object_layout(object_t oid, int pg_pool, string nspace) const;
 
   int get_pg_num(int pg_pool) const
   {
@@ -604,7 +625,7 @@ private:
   void print_osd_line(int cur, ostream *out, Formatter *f) const;
 public:
   void print(ostream& out) const;
-  void print_summary(ostream& out) const;
+  void print_summary(Formatter *f, ostream& out) const;
   void print_tree(ostream *out, Formatter *f) const;
 
   string get_flag_string() const;
@@ -620,7 +641,7 @@ WRITE_CLASS_ENCODER_FEATURES(OSDMap::Incremental)
 typedef std::tr1::shared_ptr<const OSDMap> OSDMapRef;
 
 inline ostream& operator<<(ostream& out, const OSDMap& m) {
-  m.print_summary(out);
+  m.print_summary(NULL, out);
   return out;
 }
 
